# Copyright (C) 2011-2018 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

import vim
import json
from ycm import vimsupport
from ycmd import utils
from ycm.vimsupport import memoize, GetIntValue


class SignatureHelpState:
  ACTIVE = 'ACTIVE'
  INACTIVE = 'INACTIVE'
  ACTIVE_SUPPRESSED = 'ACTIVE_SUPPRESSED'

  def __init__( self,
                popup_win_id = None,
                state = INACTIVE ):
    self.popup_win_id = popup_win_id
    self.state = state
    self.anchor = None


  def ToggleVisibility( self ):
    if self.state == 'ACTIVE':
      self.state = 'ACTIVE_SUPPRESSED'
      vim.eval( f'popup_hide( { self.popup_win_id } )' )
    elif self.state == 'ACTIVE_SUPPRESSED':
      self.state = 'ACTIVE'
      vim.eval( f'popup_show( { self.popup_win_id } )' )


  def IsActive( self ):
    if self.state in ( 'ACTIVE', 'ACTIVE_SUPPRESSED' ):
      return 'ACTIVE'
    return 'INACTIVE'


def _MakeSignatureHelpBuffer( signature_info ):
  active_parameter = int( signature_info.get( 'activeParameter', 0 ) )

  lines = []
  signatures = ( signature_info.get( 'signatures' ) or [] )

  for sig_index, signature in enumerate( signatures ):
    props = []

    sig_label = signature[ 'label' ]
    parameters = ( signature.get( 'parameters' ) or [] )
    for param_index, parameter in enumerate( parameters ):
      param_label = parameter[ 'label' ]
      begin = int( param_label[ 0 ] )
      end = int( param_label[ 1 ] )
      if param_index == active_parameter:
        props.append( {
          'col': begin + 1, # 1-based
          'length': end - begin,
          'type': 'YCM-signature-help-current-argument'
        } )

    lines.append( {
      'text': sig_label,
      'props': props
    } )

  return lines


@memoize()
def ShouldUseSignatureHelp():
  return ( vimsupport.VimHasFunctions( 'screenpos', 'pum_getpos' ) and
           vimsupport.VimSupportsPopupWindows() )


def UpdateSignatureHelp( state, signature_info ): # noqa
  if not ShouldUseSignatureHelp():
    return state

  signatures = signature_info.get( 'signatures' ) or []

  if not signatures:
    if state.popup_win_id:
      # TODO/FIXME: Should we use popup_hide() instead ?
      vim.eval( f"popup_close( { state.popup_win_id } )" )
    return SignatureHelpState( None, SignatureHelpState.INACTIVE )

  if state.state == SignatureHelpState.INACTIVE:
    state.anchor = vimsupport.CurrentLineAndColumn()

    state.state = SignatureHelpState.ACTIVE

  # Generate the buffer as a list of lines
  buf_lines = _MakeSignatureHelpBuffer( signature_info )
  screen_pos = vimsupport.ScreenPositionForLineColumnInWindow(
    vim.current.window,
    state.anchor[ 0 ] + 1,  # anchor 0-based
    state.anchor[ 1 ] + 1 ) # anchor 0-based

  # Simulate 'flip' at the screen boundaries by using screenpos and hiding the
  # signature help menu if it overlaps the completion popup (pum).
  #
  # FIXME: revert to cursor-relative positioning and the 'flip' option when that
  # is implemented (if that is indeed better).

  # By default display above the anchor
  line = int( screen_pos[ 'row' ] ) - 1 # -1 to display above the cur line
  pos = "botleft"

  cursor_line = vimsupport.CurrentLineAndColumn()[ 0 ] + 1
  if int( screen_pos[ 'row' ] ) <= len( buf_lines ):
    # No room at the top, display below
    line = int( screen_pos[ 'row' ] ) + 1
    pos = "topleft"

  # Don't allow the popup to overlap the cursor
  if ( pos == 'topleft' and
       line < cursor_line and
       line + len( buf_lines ) >= cursor_line ):
    line = 0

  # Don't allow the popup to overlap the pum
  if line > 0 and GetIntValue( 'pumvisible()' ):
    pum_line = GetIntValue( 'pum_getpos().row' ) + 1
    if pos == 'botleft' and pum_line <= line:
      line = 0
    elif ( pos == 'topleft' and
           pum_line >= line and
           pum_line < ( line + len( buf_lines ) ) ):
      line = 0

  if line <= 0:
    # Nowhere to put it so hide it
    if state.popup_win_id:
      # TODO/FIXME: Should we use popup_hide() instead ?
      vim.eval( f"popup_close( { state.popup_win_id } )" )
    return SignatureHelpState( None, SignatureHelpState.INACTIVE )

  if int( screen_pos[ 'curscol' ] ) <= 1:
    col = 1
  else:
    # -1 for padding,
    # -1 for the trigger character inserted (the anchor is set _after_ the
    # character is inserted, so we remove it).
    # FIXME: multi-byte characters would be wrong. Need to set anchor before
    # inserting the char ?
    col = int( screen_pos[ 'curscol' ] ) - 2

  if col <= 0:
    col = 1

  options = {
    "line": line,
    "col": col,
    "pos": pos,
    "wrap": 0,
    # NOTE: We *dont'* use "cursorline" here - that actually uses PMenuSel,
    # which is just too invasive for us (it's more selected item than actual
    # cursorline. So instead, we manually set 'cursorline' in the popup window
    # and enable syntax based on the current file syntax)
    "flip": 1,
    "padding": [ 0, 1, 0, 1 ], # Pad 1 char in X axis to match completion menu
    "hidden": int( state.state == SignatureHelpState.ACTIVE_SUPPRESSED )
  }

  if not state.popup_win_id:
    state.popup_win_id = GetIntValue(
      f'popup_create( { json.dumps( buf_lines ) }, '
                    f'{ json.dumps( options ) } )' )
  else:
    vim.eval( f'popup_settext( { state.popup_win_id }, '
                             f'{ json.dumps( buf_lines ) } )' )

  # Should do nothing if already visible
  vim.eval( f'popup_move( { state.popup_win_id }, { json.dumps( options ) } )' )
  if state.state == SignatureHelpState.ACTIVE:
    vim.eval( f'popup_show( { state.popup_win_id } )' )

  syntax = utils.ToUnicode( vim.current.buffer.options[ 'syntax' ] )
  active_signature = int( signature_info.get( 'activeSignature', 0 ) )
  vim.eval( f"win_execute( { state.popup_win_id }, "
            f"'set syntax={ syntax } cursorline | "
            f"call cursor( [ { active_signature + 1 }, 1 ] )' )" )

  return state



# Copyright (C) 2020, YouCompleteMe Contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.


from ycm import vimsupport
from ycm.vimsupport import GetIntValue
from ycmd import utils

import vim
import json
import typing


# FIXME/TODO: Merge this with vimsupport funcitons, added after these were
# written. It's not trivial, as those vimsupport functions are a bit fiddly.
# They also support neovim, but we don't.
def AddTextPropertyType( name, **kwargs ):
  props = {
    'highlight': 'Ignore',
    'combine': 0,
    'override': 0,
    'start_incl': 0,
    'end_incl': 0,
    'priority': 10
  }
  props.update( kwargs )

  vim.eval( f"prop_type_add( '{ vimsupport.EscapeForVim( name ) }', "
            f"               { json.dumps( props ) } )" )


def GetTextPropertyTypes( *args, **kwargs ):
  return [ utils.ToUnicode( p ) for p in vim.eval( 'prop_type_list()' ) ]


def AddTextProperty( bufnr,
                     prop_id,
                     prop_type,
                     range,
                     extra_args: dict = None ):
  props = {
    'bufnr': bufnr,
    'type': prop_type
  }
  if prop_id is not None:
    props[ 'id' ] = prop_id
  if extra_args:
    props.update( extra_args )
  if 'end' in range:
    props.update( {
      'end_lnum': range[ 'end' ][ 'line_num' ],
      'end_col':  range[ 'end' ][ 'column_num' ],
    } )
  return vim.eval( f"prop_add( { range[ 'start' ][ 'line_num' ] },"
                   f"          { range[ 'start' ][ 'column_num' ] },"
                   f"          { json.dumps( props ) } )" )


def ClearTextProperties(
  bufnr,
  prop_id = None,
  prop_types: typing.Union[ typing.List[ str ], str ] = None,
  first_line = None,
  last_line = None ):

  props = {
    'bufnr': bufnr,
    'all': 1,
  }
  if prop_id is not None:
    props[ 'id' ] = prop_id

  if prop_id is not None and prop_types is not None:
    props[ 'both' ] = 1

  def prop_remove():
    if last_line is not None:
      return GetIntValue( f"prop_remove( { json.dumps( props ) },"
                                      f" { first_line },"
                                      f" { last_line } )" )
    elif first_line is not None:
      return GetIntValue( f"prop_remove( { json.dumps( props ) },"
                                      f" { first_line } )" )
    else:
      return GetIntValue( f"prop_remove( { json.dumps( props ) } )" )

  if prop_types is None:
    return prop_remove()

  if not isinstance( prop_types, list ):
    prop_types = [ prop_types ]

  # 9.0.233 added types list to prop_remove, so use that
  if vimsupport.VimVersionAtLeast( '9.0.233' ):
    props[ 'types' ] = prop_types
    return prop_remove()

  # Older versions we have to run prop_remove for each type
  removed = 0
  for prop_type in prop_types:
    props[ 'type' ] = prop_type
    removed += prop_remove()
  return removed



# Copyright (C) 2016-2017 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.tests.test_utils import MockVimModule
MockVimModule()

from hamcrest import assert_that
from unittest import TestCase
from ycm.paths import _EndsWithPython


def EndsWithPython_Good( path ):
  assert_that( _EndsWithPython( path ),
              f'Path { path } does not end with a Python name.' )


def EndsWithPython_Bad( path ):
  assert_that( not _EndsWithPython( path ),
              f'Path { path } does end with a Python name.' )


class PathTest( TestCase ):
  def test_EndsWithPython_Python3Paths( self ):
    for path in [
      'python3',
      '/usr/bin/python3.6',
      '/home/user/.pyenv/shims/python3.6',
      r'C:\Python36\python.exe'
    ]:
      with self.subTest( path = path ):
        EndsWithPython_Good( path )


  def test_EndsWithPython_BadPaths( self ):
    for path in [
      None,
      '',
      '/opt/local/bin/vim',
      r'C:\Program Files\Vim\vim74\gvim.exe',
      '/usr/bin/python2.7',
      '/home/user/.pyenv/shims/python3.2',
    ]:
      with self.subTest( path = path ):
        EndsWithPython_Bad( path )



# Copyright (C) 2015-2018 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.tests.test_utils import ( CurrentWorkingDirectory, ExtendedMock,
                                   MockVimBuffers, MockVimModule, VimBuffer,
                                   VimSign )
MockVimModule()

import contextlib
import os

from ycm.tests import ( PathToTestFile, test_utils, YouCompleteMeInstance,
                        WaitUntilReady )
from ycmd.responses import ( BuildDiagnosticData, Diagnostic, Location, Range,
                             UnknownExtraConf, ServerError )

from hamcrest import ( assert_that, contains_exactly, empty, equal_to,
                       has_entries, has_entry, has_item, has_items, has_key,
                       is_not )
from unittest import TestCase
from unittest.mock import call, MagicMock, patch


def PresentDialog_Confirm_Call( message ):
  """Return a mock.call object for a call to vimsupport.PresentDialog, as called
  why vimsupport.Confirm with the supplied confirmation message"""
  return call( message, [ 'Ok', 'Cancel' ] )


@contextlib.contextmanager
def MockArbitraryBuffer( filetype ):
  """Used via the with statement, set up a single buffer with an arbitrary name
  and no contents. Its filetype is set to the supplied filetype."""

  # Arbitrary, but valid, single buffer open.
  current_buffer = VimBuffer( os.path.realpath( 'TEST_BUFFER' ),
                              filetype = filetype )

  with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
    yield


@contextlib.contextmanager
def MockEventNotification( response_method, native_filetype_completer = True ):
  """Mock out the EventNotification client request object, replacing the
  Response handler's JsonFromFuture with the supplied |response_method|.
  Additionally mock out YouCompleteMe's FiletypeCompleterExistsForFiletype
  method to return the supplied |native_filetype_completer| parameter, rather
  than querying the server"""

  # We don't want the event to actually be sent to the server, just have it
  # return success
  with patch( 'ycm.client.event_notification.EventNotification.'
              'PostDataToHandlerAsync',
              return_value = MagicMock( return_value=True ) ):

    # We set up a fake a Response (as called by EventNotification.Response)
    # which calls the supplied callback method. Generally this callback just
    # raises an apropriate exception, otherwise it would have to return a mock
    # future object.
    with patch( 'ycm.client.base_request._JsonFromFuture',
                side_effect = response_method ):

      # Filetype available information comes from the server, so rather than
      # relying on that request, we mock out the check. The caller decides if
      # filetype completion is available
      with patch(
        'ycm.youcompleteme.YouCompleteMe.FiletypeCompleterExistsForFiletype',
        return_value = native_filetype_completer ):

        yield


def _Check_FileReadyToParse_Diagnostic_Error( ycm ):
  # Tests Vim sign placement and error/warning count python API
  # when one error is returned.
  def DiagnosticResponse( *args ):
    start = Location( 1, 2, 'TEST_BUFFER' )
    end = Location( 1, 4, 'TEST_BUFFER' )
    extent = Range( start, end )
    diagnostic = Diagnostic( [], start, extent, 'expected ;', 'ERROR' )
    return [ BuildDiagnosticData( diagnostic ) ]

  with MockArbitraryBuffer( 'cpp' ):
    with MockEventNotification( DiagnosticResponse ):
      ycm.OnFileReadyToParse()
      assert_that( ycm.FileParseRequestReady() )
      ycm.HandleFileParseRequest()
      assert_that(
        test_utils.VIM_SIGNS,
        contains_exactly(
          VimSign( 1, 'YcmError', 1 )
        )
      )
      assert_that( ycm.GetErrorCount(), equal_to( 1 ) )
      assert_that( ycm.GetWarningCount(), equal_to( 0 ) )

      # Consequent calls to HandleFileParseRequest shouldn't mess with
      # existing diagnostics, when there is no new parse request.
      ycm.HandleFileParseRequest()
      assert_that(
        test_utils.VIM_SIGNS,
        contains_exactly(
          VimSign( 1, 'YcmError', 1 )
        )
      )
      assert_that( ycm.GetErrorCount(), equal_to( 1 ) )
      assert_that( ycm.GetWarningCount(), equal_to( 0 ) )

      assert_that( not ycm.ShouldResendFileParseRequest() )

      # New identical requests should result in the same diagnostics.
      ycm.OnFileReadyToParse()
      assert_that( ycm.FileParseRequestReady() )
      ycm.HandleFileParseRequest()
      assert_that(
        test_utils.VIM_SIGNS,
        contains_exactly(
          VimSign( 1, 'YcmError', 1 )
        )
      )
      assert_that( ycm.GetErrorCount(), equal_to( 1 ) )
      assert_that( ycm.GetWarningCount(), equal_to( 0 ) )

      assert_that( not ycm.ShouldResendFileParseRequest() )


def _Check_FileReadyToParse_Diagnostic_Warning( ycm ):
  # Tests Vim sign placement/unplacement and error/warning count python API
  # when one warning is returned.
  # Should be called after _Check_FileReadyToParse_Diagnostic_Error
  def DiagnosticResponse( *args ):
    start = Location( 2, 2, 'TEST_BUFFER' )
    end = Location( 2, 4, 'TEST_BUFFER' )
    extent = Range( start, end )
    diagnostic = Diagnostic( [], start, extent, 'cast', 'WARNING' )
    return [ BuildDiagnosticData( diagnostic ) ]

  with MockArbitraryBuffer( 'cpp' ):
    with MockEventNotification( DiagnosticResponse ):
      ycm.OnFileReadyToParse()
      assert_that( ycm.FileParseRequestReady() )
      ycm.HandleFileParseRequest()
      assert_that(
        test_utils.VIM_SIGNS,
        contains_exactly(
          VimSign( 2, 'YcmWarning', 1 )
        )
      )
      assert_that( ycm.GetErrorCount(), equal_to( 0 ) )
      assert_that( ycm.GetWarningCount(), equal_to( 1 ) )

      # Consequent calls to HandleFileParseRequest shouldn't mess with
      # existing diagnostics, when there is no new parse request.
      ycm.HandleFileParseRequest()
      assert_that(
        test_utils.VIM_SIGNS,
        contains_exactly(
          VimSign( 2, 'YcmWarning', 1 )
        )
      )
      assert_that( ycm.GetErrorCount(), equal_to( 0 ) )
      assert_that( ycm.GetWarningCount(), equal_to( 1 ) )

      assert_that( not ycm.ShouldResendFileParseRequest() )


def _Check_FileReadyToParse_Diagnostic_Clean( ycm ):
  # Tests Vim sign unplacement and error/warning count python API
  # when there are no errors/warnings left.
  # Should be called after _Check_FileReadyToParse_Diagnostic_Warning
  with MockArbitraryBuffer( 'cpp' ):
    with MockEventNotification( MagicMock( return_value = [] ) ):
      ycm.OnFileReadyToParse()
      ycm.HandleFileParseRequest()
      assert_that(
        test_utils.VIM_SIGNS,
        empty()
      )
      assert_that( ycm.GetErrorCount(), equal_to( 0 ) )
      assert_that( ycm.GetWarningCount(), equal_to( 0 ) )
      assert_that( not ycm.ShouldResendFileParseRequest() )


class EventNotificationTest( TestCase ):
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  @YouCompleteMeInstance()
  def test_EventNotification_FileReadyToParse_NonDiagnostic_Error(
      self, ycm, post_vim_message ):

    # This test validates the behaviour of YouCompleteMe.HandleFileParseRequest
    # in combination with YouCompleteMe.OnFileReadyToParse when the completer
    # raises an exception handling FileReadyToParse event notification
    ERROR_TEXT = 'Some completer response text'

    def ErrorResponse( *args ):
      raise ServerError( ERROR_TEXT )

    with MockArbitraryBuffer( 'some_filetype' ):
      with MockEventNotification( ErrorResponse ):
        ycm.OnFileReadyToParse()
        assert_that( ycm.FileParseRequestReady() )
        ycm.HandleFileParseRequest()

        # The first call raises a warning
        post_vim_message.assert_has_exact_calls( [
          call( ERROR_TEXT, truncate = True )
        ] )

        # Subsequent calls don't re-raise the warning
        ycm.HandleFileParseRequest()
        post_vim_message.assert_has_exact_calls( [
          call( ERROR_TEXT, truncate = True )
        ] )

        assert_that( not ycm.ShouldResendFileParseRequest() )

        # But it does if a subsequent event raises again
        ycm.OnFileReadyToParse()
        assert_that( ycm.FileParseRequestReady() )
        ycm.HandleFileParseRequest()
        post_vim_message.assert_has_exact_calls( [
          call( ERROR_TEXT, truncate = True ),
          call( ERROR_TEXT, truncate = True )
        ] )

        assert_that( not ycm.ShouldResendFileParseRequest() )


  @YouCompleteMeInstance()
  def test_EventNotification_FileReadyToParse_NonDiagnostic_Error_NonNative(
    self, ycm ):

    test_utils.VIM_MATCHES = []
    test_utils.VIM_SIGNS = []

    with MockArbitraryBuffer( 'some_filetype' ):
      with MockEventNotification( None, False ):
        ycm.OnFileReadyToParse()
        ycm.HandleFileParseRequest()
        assert_that( test_utils.VIM_MATCHES, empty() )
        assert_that( test_utils.VIM_SIGNS, empty() )
        assert_that( not ycm.ShouldResendFileParseRequest() )


  @YouCompleteMeInstance()
  def test_EventNotification_FileReadyToParse_NonDiagnostic_ConfirmExtraConf(
      self, ycm ):

    # This test validates the behaviour of YouCompleteMe.HandleFileParseRequest
    # in combination with YouCompleteMe.OnFileReadyToParse when the completer
    # raises the (special) UnknownExtraConf exception
    FILE_NAME = 'a_file'
    MESSAGE = ( 'Found ' + FILE_NAME + '. Load? \n\n(Question can be '
                'turned off with options, see YCM docs)' )

    def UnknownExtraConfResponse( *args ):
      raise UnknownExtraConf( FILE_NAME )

    with patch( 'ycm.client.base_request.BaseRequest.PostDataToHandler',
                new_callable = ExtendedMock ) as post_data_to_handler:
      with MockArbitraryBuffer( 'some_filetype' ):
        with MockEventNotification( UnknownExtraConfResponse ):

          # When the user accepts the extra conf, we load it
          with patch( 'ycm.vimsupport.PresentDialog',
                      return_value = 0,
                      new_callable = ExtendedMock ) as present_dialog:
            ycm.OnFileReadyToParse()
            assert_that( ycm.FileParseRequestReady() )
            ycm.HandleFileParseRequest()

            present_dialog.assert_has_exact_calls( [
              PresentDialog_Confirm_Call( MESSAGE ),
            ] )
            post_data_to_handler.assert_has_exact_calls( [
              call( { 'filepath': FILE_NAME }, 'load_extra_conf_file' )
            ] )

            # Subsequent calls don't re-raise the warning
            ycm.HandleFileParseRequest()

            present_dialog.assert_has_exact_calls( [
              PresentDialog_Confirm_Call( MESSAGE )
            ] )
            post_data_to_handler.assert_has_exact_calls( [
              call( { 'filepath': FILE_NAME }, 'load_extra_conf_file' )
            ] )

            assert_that( ycm.ShouldResendFileParseRequest() )

            # But it does if a subsequent event raises again
            ycm.OnFileReadyToParse()
            assert_that( ycm.FileParseRequestReady() )
            ycm.HandleFileParseRequest()

            present_dialog.assert_has_exact_calls( [
              PresentDialog_Confirm_Call( MESSAGE ),
              PresentDialog_Confirm_Call( MESSAGE ),
            ] )
            post_data_to_handler.assert_has_exact_calls( [
              call( { 'filepath': FILE_NAME }, 'load_extra_conf_file' ),
              call( { 'filepath': FILE_NAME }, 'load_extra_conf_file' )
            ] )

            assert_that( ycm.ShouldResendFileParseRequest() )

          post_data_to_handler.reset_mock()

          # When the user rejects the extra conf, we reject it
          with patch( 'ycm.vimsupport.PresentDialog',
                      return_value = 1,
                      new_callable = ExtendedMock ) as present_dialog:
            ycm.OnFileReadyToParse()
            assert_that( ycm.FileParseRequestReady() )
            ycm.HandleFileParseRequest()

            present_dialog.assert_has_exact_calls( [
              PresentDialog_Confirm_Call( MESSAGE ),
            ] )
            post_data_to_handler.assert_has_exact_calls( [
              call( { 'filepath': FILE_NAME }, 'ignore_extra_conf_file' )
            ] )

            # Subsequent calls don't re-raise the warning
            ycm.HandleFileParseRequest()

            present_dialog.assert_has_exact_calls( [
              PresentDialog_Confirm_Call( MESSAGE )
            ] )
            post_data_to_handler.assert_has_exact_calls( [
              call( { 'filepath': FILE_NAME }, 'ignore_extra_conf_file' )
            ] )

            assert_that( ycm.ShouldResendFileParseRequest() )

            # But it does if a subsequent event raises again
            ycm.OnFileReadyToParse()
            assert_that( ycm.FileParseRequestReady() )
            ycm.HandleFileParseRequest()

            present_dialog.assert_has_exact_calls( [
              PresentDialog_Confirm_Call( MESSAGE ),
              PresentDialog_Confirm_Call( MESSAGE ),
            ] )
            post_data_to_handler.assert_has_exact_calls( [
              call( { 'filepath': FILE_NAME }, 'ignore_extra_conf_file' ),
              call( { 'filepath': FILE_NAME }, 'ignore_extra_conf_file' )
            ] )

            assert_that( ycm.ShouldResendFileParseRequest() )


  @YouCompleteMeInstance()
  def test_EventNotification_FileReadyToParse_Diagnostic_Error_Native(
      self, ycm ):
    test_utils.VIM_SIGNS = []

    _Check_FileReadyToParse_Diagnostic_Error( ycm )
    _Check_FileReadyToParse_Diagnostic_Warning( ycm )
    _Check_FileReadyToParse_Diagnostic_Clean( ycm )


  @patch( 'ycm.youcompleteme.YouCompleteMe._AddUltiSnipsDataIfNeeded' )
  @YouCompleteMeInstance( { 'g:ycm_collect_identifiers_from_tags_files': 1 } )
  def test_EventNotification_FileReadyToParse_TagFiles_UnicodeWorkingDirectory(
      self, ycm, *args ):
    unicode_dir = PathToTestFile( 'uni¢od€' )
    current_buffer_file = PathToTestFile( 'uni¢𐍈d€', 'current_buffer' )
    current_buffer = VimBuffer( name = current_buffer_file,
                                contents = [ 'current_buffer_contents' ],
                                filetype = 'some_filetype' )

    with patch( 'ycm.client.event_notification.EventNotification.'
                'PostDataToHandlerAsync' ) as post_data_to_handler_async:
      with CurrentWorkingDirectory( unicode_dir ):
        with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 5 ) ):
          ycm.OnFileReadyToParse()

      assert_that(
        # Positional arguments passed to PostDataToHandlerAsync.
        post_data_to_handler_async.call_args[ 0 ],
        contains_exactly(
          has_entries( {
            'filepath': current_buffer_file,
            'line_num': 1,
            'column_num': 6,
            'file_data': has_entries( {
              current_buffer_file: has_entries( {
                'contents': 'current_buffer_contents\n',
                'filetypes': [ 'some_filetype' ]
              } )
            } ),
            'event_name': 'FileReadyToParse',
            'tag_files': has_item( PathToTestFile( 'uni¢od€', 'tags' ) )
          } ),
          'event_notification'
        )
      )


  @patch( 'ycm.youcompleteme.YouCompleteMe._AddUltiSnipsDataIfNeeded' )
  @YouCompleteMeInstance()
  def test_EventNotification_BufferVisit_BuildRequestForCurrentAndUnsavedBuffers( # noqa
      self, ycm, *args ):

    current_buffer_file = os.path.realpath( 'current_buffer' )
    current_buffer = VimBuffer( name = current_buffer_file,
                                number = 1,
                                contents = [ 'current_buffer_contents' ],
                                filetype = 'some_filetype',
                                modified = False )
    modified_buffer_file = os.path.realpath( 'modified_buffer' )
    modified_buffer = VimBuffer( name = modified_buffer_file,
                                 number = 2,
                                 contents = [ 'modified_buffer_contents' ],
                                 filetype = 'some_filetype',
                                 modified = True )

    unmodified_buffer_file = os.path.realpath( 'unmodified_buffer' )
    unmodified_buffer = VimBuffer( name = unmodified_buffer_file,
                                   number = 3,
                                   contents = [ 'unmodified_buffer_contents' ],
                                   filetype = 'some_filetype',
                                   modified = False )

    with patch( 'ycm.client.event_notification.EventNotification.'
                'PostDataToHandlerAsync' ) as post_data_to_handler_async:
      with MockVimBuffers( [ current_buffer,
                             modified_buffer,
                             unmodified_buffer ],
                           [ current_buffer ],
                           ( 1, 5 ) ):
        ycm.OnBufferVisit()

      assert_that(
        # Positional arguments passed to PostDataToHandlerAsync.
        post_data_to_handler_async.call_args[ 0 ],
        contains_exactly(
          has_entries( {
            'filepath': current_buffer_file,
            'line_num': 1,
            'column_num': 6,
            'file_data': has_entries( {
              current_buffer_file: has_entries( {
                'contents': 'current_buffer_contents\n',
                'filetypes': [ 'some_filetype' ]
              } ),
              modified_buffer_file: has_entries( {
                'contents': 'modified_buffer_contents\n',
                'filetypes': [ 'some_filetype' ]
              } )
            } ),
            'event_name': 'BufferVisit'
          } ),
          'event_notification'
        )
      )


  @YouCompleteMeInstance()
  def test_EventNotification_BufferUnload_BuildRequestForDeletedAndUnsavedBuffers( # noqa
      self, ycm ):
    current_buffer_file = os.path.realpath( 'current_βuffer' )
    current_buffer = VimBuffer( name = current_buffer_file,
                                number = 1,
                                contents = [ 'current_buffer_contents' ],
                                filetype = 'some_filetype',
                                modified = True )

    deleted_buffer_file = os.path.realpath( 'deleted_βuffer' )
    deleted_buffer = VimBuffer( name = deleted_buffer_file,
                                number = 2,
                                contents = [ 'deleted_buffer_contents' ],
                                filetype = 'some_filetype',
                                modified = False )

    with patch( 'ycm.client.event_notification.EventNotification.'
                'PostDataToHandlerAsync' ) as post_data_to_handler_async:
      with MockVimBuffers( [ current_buffer, deleted_buffer ],
                           [ current_buffer ] ):
        ycm.OnBufferUnload( deleted_buffer.number )

    assert_that(
      # Positional arguments passed to PostDataToHandlerAsync.
      post_data_to_handler_async.call_args[ 0 ],
      contains_exactly(
        has_entries( {
          'filepath': deleted_buffer_file,
          'line_num': 1,
          'column_num': 1,
          'file_data': has_entries( {
            current_buffer_file: has_entries( {
              'contents': 'current_buffer_contents\n',
              'filetypes': [ 'some_filetype' ]
            } ),
            deleted_buffer_file: has_entries( {
              'contents': 'deleted_buffer_contents\n',
              'filetypes': [ 'some_filetype' ]
            } )
          } ),
          'event_name': 'BufferUnload'
        } ),
        'event_notification'
      )
    )


  @patch( 'ycm.vimsupport.CaptureVimCommand', return_value = """
fooGroup xxx foo bar
             links to Statement""" )
  @YouCompleteMeInstance( { 'g:ycm_seed_identifiers_with_syntax': 1 } )
  def test_EventNotification_FileReadyToParse_SyntaxKeywords_SeedWithCache(
      self, ycm, *args ):

    current_buffer = VimBuffer( name = 'current_buffer',
                                filetype = 'some_filetype' )

    with patch( 'ycm.client.event_notification.EventNotification.'
                'PostDataToHandlerAsync' ) as post_data_to_handler_async:
      with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
        ycm.OnFileReadyToParse()
        assert_that(
          # Positional arguments passed to PostDataToHandlerAsync.
          post_data_to_handler_async.call_args[ 0 ],
          contains_exactly(
            has_entry( 'syntax_keywords', has_items( 'foo', 'bar' ) ),
            'event_notification'
          )
        )

        # Do not send again syntax keywords in subsequent requests.
        ycm.OnFileReadyToParse()
        assert_that(
          # Positional arguments passed to PostDataToHandlerAsync.
          post_data_to_handler_async.call_args[ 0 ],
          contains_exactly(
            is_not( has_key( 'syntax_keywords' ) ),
            'event_notification'
          )
        )


  @patch( 'ycm.vimsupport.CaptureVimCommand', return_value = """
fooGroup xxx foo bar
             links to Statement""" )
  @YouCompleteMeInstance( { 'g:ycm_seed_identifiers_with_syntax': 1 } )
  def test_EventNotification_FileReadyToParse_SyntaxKeywords_ClearCacheIfRestart( # noqa
      self, ycm, *args ):

    current_buffer = VimBuffer( name = 'current_buffer',
                                filetype = 'some_filetype' )

    with patch( 'ycm.client.event_notification.EventNotification.'
                'PostDataToHandlerAsync' ) as post_data_to_handler_async:
      with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
        ycm.OnFileReadyToParse()
        assert_that(
          # Positional arguments passed to PostDataToHandlerAsync.
          post_data_to_handler_async.call_args[ 0 ],
          contains_exactly(
            has_entry( 'syntax_keywords', has_items( 'foo', 'bar' ) ),
            'event_notification'
          )
        )

        # Send again the syntax keywords after restarting the server.
        ycm.RestartServer()
        WaitUntilReady()
        ycm.OnFileReadyToParse()
        assert_that(
          # Positional arguments passed to PostDataToHandlerAsync.
          post_data_to_handler_async.call_args[ 0 ],
          contains_exactly(
            has_entry( 'syntax_keywords', has_items( 'foo', 'bar' ) ),
            'event_notification'
          )
        )



# Copyright (C) 2016 YouCompleteMe Contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.tests.test_utils import ExtendedMock, MockVimModule
MockVimModule()

import json
from hamcrest import assert_that
from unittest import TestCase
from unittest.mock import patch, call
from ycm.client.command_request import CommandRequest


def GoToTest( command, response ):
  with patch( 'ycm.vimsupport.JumpToLocation' ) as jump_to_location:
    request = CommandRequest( [ command ] )
    request._response = response
    request.RunPostCommandActionsIfNeeded( 'rightbelow' )
    jump_to_location.assert_called_with(
        response[ 'filepath' ],
        response[ 'line_num' ],
        response[ 'column_num' ],
        'rightbelow',
        'same-buffer' )


def GoToListTest( command, response ):
  # Note: the detail of these called are tested by
  # GoToResponse_QuickFix_test, so here we just check that the right call is
  # made
  with patch( 'ycm.vimsupport.SetQuickFixList' ) as set_qf_list:
    with patch( 'ycm.vimsupport.OpenQuickFixList' ) as open_qf_list:
      request = CommandRequest( [ command ] )
      request._response = response
      request.RunPostCommandActionsIfNeeded( 'tab' )
      assert_that( set_qf_list.called )
      assert_that( open_qf_list.called )


BASIC_GOTO = {
  'filepath': 'test',
  'line_num': 10,
  'column_num': 100,
}


BASIC_FIXIT = {
  'fixits': [ {
    'resolve': False,
    'chunks': [ {
      'dummy chunk contents': True
    } ]
  } ]
}
BASIC_FIXIT_CHUNKS = BASIC_FIXIT[ 'fixits' ][ 0 ][ 'chunks' ]

MULTI_FIXIT = {
  'fixits': [ {
    'text': 'first',
    'resolve': False,
    'chunks': [ {
      'dummy chunk contents': True
    } ]
  }, {
    'text': 'second',
    'resolve': False,
    'chunks': [ {
      'dummy chunk contents': False
    } ]
  } ]
}
MULTI_FIXIT_FIRST_CHUNKS = MULTI_FIXIT[ 'fixits' ][ 0 ][ 'chunks' ]
MULTI_FIXIT_SECOND_CHUNKS = MULTI_FIXIT[ 'fixits' ][ 1 ][ 'chunks' ]


class GoToResponse_QuickFixTest( TestCase ):
  """This class tests the generation of QuickFix lists for GoTo responses which
  return multiple locations, such as the Python completer and JavaScript
  completer. It mostly proves that we use 1-based indexing for the column
  number."""

  def setUp( self ):
    self._request = CommandRequest( [ 'GoToTest' ] )


  def tearDown( self ):
    self._request = None


  def test_GoTo_EmptyList( self ):
    self._CheckGoToList( [], [] )


  def test_GoTo_SingleItem_List( self ):
    self._CheckGoToList( [ {
      'filepath':     'dummy_file',
      'line_num':     10,
      'column_num':   1,
      'description': 'this is some text',
    } ], [ {
      'filename':    'dummy_file',
      'text':        'this is some text',
      'lnum':        10,
      'col':         1
    } ] )


  def test_GoTo_MultiItem_List( self ):
    self._CheckGoToList( [ {
      'filepath':     'dummy_file',
      'line_num':     10,
      'column_num':   1,
      'description': 'this is some other text',
    }, {
      'filepath':     'dummy_file2',
      'line_num':     1,
      'column_num':   21,
      'description': 'this is some text',
    } ], [ {
      'filename':    'dummy_file',
      'text':        'this is some other text',
      'lnum':        10,
      'col':         1
    }, {
      'filename':    'dummy_file2',
      'text':        'this is some text',
      'lnum':        1,
      'col':         21
    } ] )


  @patch( 'ycm.vimsupport.VariableExists', return_value = True )
  @patch( 'ycm.vimsupport.SetFittingHeightForCurrentWindow' )
  @patch( 'vim.command', new_callable = ExtendedMock )
  @patch( 'vim.eval', new_callable = ExtendedMock )
  def _CheckGoToList( self,
                      completer_response,
                      expected_qf_list,
                      vim_eval,
                      vim_command,
                      set_fitting_height,
                      variable_exists ):
    self._request._response = completer_response

    self._request.RunPostCommandActionsIfNeeded( 'aboveleft' )

    vim_eval.assert_has_exact_calls( [
      call( f'setqflist( { json.dumps( expected_qf_list ) } )' )
    ] )
    vim_command.assert_has_exact_calls( [
      call( 'botright copen' ),
      call( 'augroup ycmquickfix' ),
      call( 'autocmd! * <buffer>' ),
      call( 'autocmd WinLeave <buffer> '
            'if bufnr( "%" ) == expand( "<abuf>" ) | q | endif '
            '| autocmd! ycmquickfix' ),
      call( 'augroup END' ),
      call( 'doautocmd User YcmQuickFixOpened' )
    ] )
    set_fitting_height.assert_called_once_with()


class Response_Detection_Test( TestCase ):

  def test_BasicResponse( self ):
    def _BasicResponseTest( command, response ):
      with patch( 'vim.command' ) as vim_command:
        request = CommandRequest( [ command ] )
        request._response = response
        request.RunPostCommandActionsIfNeeded( 'belowright' )
        vim_command.assert_called_with( f"echo '{ response }'" )

    for command, response in [
        [ 'AnythingYouLike',        True ],
        [ 'GoToEvenWorks',          10 ],
        [ 'FixItWorks',             'String!' ],
        [ 'and8434fd andy garbag!', 10.3 ],
      ]:
      with self.subTest( command = command, response = response ):
        _BasicResponseTest( command, response )

  def test_FixIt_Response_Empty( self ):
    # Ensures we recognise and handle fixit responses which indicate that there
    # are no fixits available
    def EmptyFixItTest( command ):
      with patch( 'ycm.vimsupport.ReplaceChunks' ) as replace_chunks:
        with patch( 'ycm.vimsupport.PostVimMessage' ) as post_vim_message:
          request = CommandRequest( [ command ] )
          request._response = {
            'fixits': []
          }
          request.RunPostCommandActionsIfNeeded( 'botright' )

          post_vim_message.assert_called_with(
            'No fixits found for current line', warning = False )
          replace_chunks.assert_not_called()

    for command in [ 'FixIt', 'Refactor', 'GoToHell', 'any_old_garbade!!!21' ]:
      with self.subTest( command = command ):
        EmptyFixItTest( command )



  def test_FixIt_Response( self ):
    # Ensures we recognise and handle fixit responses with some dummy chunk data
    def FixItTest( command, response, chunks, selection, silent ):
      with patch( 'ycm.vimsupport.ReplaceChunks' ) as replace_chunks:
        with patch( 'ycm.vimsupport.PostVimMessage' ) as post_vim_message:
          with patch( 'ycm.vimsupport.SelectFromList',
                      return_value = selection ):
            request = CommandRequest( [ command ] )
            request._response = response
            request.RunPostCommandActionsIfNeeded( 'leftabove' )

            replace_chunks.assert_called_with( chunks, silent = silent )
            post_vim_message.assert_not_called()


    for command, response, chunks, selection, silent in [
      [ 'AnythingYouLike',
        BASIC_FIXIT,  BASIC_FIXIT_CHUNKS,        0, False ],
      [ 'GoToEvenWorks',
        BASIC_FIXIT,  BASIC_FIXIT_CHUNKS,        0, False ],
      [ 'FixItWorks',
        BASIC_FIXIT,  BASIC_FIXIT_CHUNKS,        0, False ],
      [ 'and8434fd andy garbag!',
        BASIC_FIXIT,  BASIC_FIXIT_CHUNKS,        0, False ],
      [ 'Format',
        BASIC_FIXIT,  BASIC_FIXIT_CHUNKS,        0, True ],
      [ 'select from multiple 1',
        MULTI_FIXIT,  MULTI_FIXIT_FIRST_CHUNKS,  0, False ],
      [ 'select from multiple 2',
        MULTI_FIXIT,  MULTI_FIXIT_SECOND_CHUNKS, 1, False ],
    ]:
      with self.subTest( command = command,
                                   response = response,
                                   chunks = chunks,
                                   selection = selection,
                                   silent = silent ):
        FixItTest( command, response, chunks, selection, silent )


  def test_Message_Response( self ):
    # Ensures we correctly recognise and handle responses with a message to show
    # to the user

    def MessageTest( command, message ):
      with patch( 'ycm.vimsupport.PostVimMessage' ) as post_vim_message:
        request = CommandRequest( [ command ] )
        request._response = { 'message': message }
        request.RunPostCommandActionsIfNeeded( 'rightbelow' )
        post_vim_message.assert_called_with( message, warning = False )

    for command, message in [
      [ '___________', 'This is a message' ],
      [ '',            'this is also a message' ],
      [ 'GetType',     'std::string' ],
    ]:
      with self.subTest( command = command, message = message ):
        MessageTest( command, message )


  def test_Detailed_Info( self ):
    # Ensures we correctly detect and handle detailed_info responses which are
    # used to display information in the preview window

    def DetailedInfoTest( command, info ):
      with patch( 'ycm.vimsupport.WriteToPreviewWindow' ) as write_to_preview:
        request = CommandRequest( [ command ] )
        request._response = { 'detailed_info': info }
        request.RunPostCommandActionsIfNeeded( 'topleft' )
        write_to_preview.assert_called_with( info )

    for command, info in [
      [ '___________', 'This is a message' ],
      [ '',            'this is also a message' ],
      [ 'GetDoc',      'std::string\netc\netc' ],
    ]:
      with self.subTest( command = command, info = info ):
        DetailedInfoTest( command, info )


  def test_GoTo_Single( self ):
    for test, command, response in [
      [ GoToTest,     'AnythingYouLike', BASIC_GOTO ],
      [ GoToTest,     'GoTo',            BASIC_GOTO ],
      [ GoToTest,     'FindAThing',      BASIC_GOTO ],
      [ GoToTest,     'FixItGoto',       BASIC_GOTO ],
      [ GoToListTest, 'AnythingYouLike', [ BASIC_GOTO ] ],
      [ GoToListTest, 'GoTo',            [] ],
      [ GoToListTest, 'FixItGoto',       [ BASIC_GOTO, BASIC_GOTO ] ],
    ]:
      with self.subTest( test = test, command = command, response = response ):
        test( command, response )



# Copyright (C) 2013-2018 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
from ycm import vimsupport
from ycm.diagnostic_filter import DiagnosticFilter, CompileLevel
from ycm import text_properties as tp
import vim
YCM_VIM_PROPERTY_ID = 1


class DiagnosticInterface:
  def __init__( self, bufnr, user_options ):
    self._bufnr = bufnr
    self._user_options = user_options
    self._diagnostics = []
    self._diag_filter = DiagnosticFilter.CreateFromOptions( user_options )
    # Line and column numbers are 1-based
    self._line_to_diags = defaultdict( list )
    self._previous_diag_line_number = -1
    self._diag_message_needs_clearing = False


  def ShouldUpdateDiagnosticsUINow( self ):
    return ( self._user_options[ 'update_diagnostics_in_insert_mode' ] or
             'i' not in vim.eval( 'mode()' ) )


  def OnCursorMoved( self ):
    if self._user_options[ 'echo_current_diagnostic' ]:
      line, _ = vimsupport.CurrentLineAndColumn()
      line += 1  # Convert to 1-based
      if ( not self.ShouldUpdateDiagnosticsUINow() and
           self._diag_message_needs_clearing ):
        # Clear any previously echo'd diagnostic in insert mode
        self._EchoDiagnosticText( line, None, None )
      elif line != self._previous_diag_line_number:
        self._EchoDiagnosticForLine( line )


  def GetErrorCount( self ):
    return self._DiagnosticsCount( _DiagnosticIsError )


  def GetWarningCount( self ):
    return self._DiagnosticsCount( _DiagnosticIsWarning )


  def PopulateLocationList( self, open_on_edit = False ):
    # Do nothing if loc list is already populated by diag_interface
    if not self._user_options[ 'always_populate_location_list' ]:
      self._UpdateLocationLists( open_on_edit )
    return bool( self._diagnostics )


  def UpdateWithNewDiagnostics( self, diags, open_on_edit = False ):
    self._diagnostics = [ _NormalizeDiagnostic( x ) for x in
                            self._ApplyDiagnosticFilter( diags ) ]
    self._ConvertDiagListToDict()

    if self.ShouldUpdateDiagnosticsUINow():
      self.RefreshDiagnosticsUI( open_on_edit )


  def RefreshDiagnosticsUI( self, open_on_edit = False ):
    if self._user_options[ 'echo_current_diagnostic' ]:
      self._EchoDiagnostic()

    if self._user_options[ 'enable_diagnostic_signs' ]:
      self._UpdateSigns()

    self.UpdateMatches()

    if self._user_options[ 'always_populate_location_list' ]:
      self._UpdateLocationLists( open_on_edit )


  def DiagnosticsForLine( self, line_number ):
    return self._line_to_diags[ line_number ]


  def _ApplyDiagnosticFilter( self, diags ):
    filetypes = vimsupport.GetBufferFiletypes( self._bufnr )
    diag_filter = self._diag_filter.SubsetForTypes( filetypes )
    return filter( diag_filter.IsAllowed, diags )


  def _EchoDiagnostic( self ):
    line, _ = vimsupport.CurrentLineAndColumn()
    line += 1  # Convert to 1-based
    self._EchoDiagnosticForLine( line )


  def _EchoDiagnosticForLine( self, line_num ):
    self._previous_diag_line_number = line_num

    diags = self._line_to_diags[ line_num ]
    text = None
    first_diag = None
    if diags:
      first_diag = diags[ 0 ]
      text = first_diag[ 'text' ]
      if first_diag.get( 'fixit_available', False ):
        text += ' (FixIt)'

    self._EchoDiagnosticText( line_num, first_diag, text )


  def _EchoDiagnosticText( self, line_num, first_diag, text ):
    if ( vimsupport.VimSupportsVirtualText() and
         self._user_options[ 'echo_current_diagnostic' ] == 'virtual-text' ):
      if self._diag_message_needs_clearing:
        # Clear any previous diag echo
        tp.ClearTextProperties( self._bufnr,
                                prop_types = [ 'YcmVirtDiagPadding',
                                               'YcmVirtDiagError',
                                               'YcmVirtDiagWarning' ] )
        self._diag_message_needs_clearing = False

      if not text:
        return

      def MakeVritualTextProperty( prop_type, text, position='after' ):
        vimsupport.AddTextProperty( self._bufnr,
                                    line_num,
                                    0,
                                    prop_type,
                                    {
                                      'text': text,
                                      'text_align': position,
                                      'text_wrap': 'wrap'
                                    } )

      if vim.options[ 'ambiwidth' ] != 'double':
        marker = '⚠'
      else:
        marker = '>'

      MakeVritualTextProperty(
        'YcmVirtDiagPadding',
        ' ' * vim.buffers[ self._bufnr ].options[ 'shiftwidth' ] ),
      MakeVritualTextProperty(
        'YcmVirtDiagError' if _DiagnosticIsError( first_diag )
                       else 'YcmVirtDiagWarning',
        marker + ' ' + [ line for line in text.splitlines() if line ][ 0 ] )
    else:
      if not text:
        if self._diag_message_needs_clearing:
          # Clear any previous diag echo
          vimsupport.PostVimMessage( '', warning = False )
          self._diag_message_needs_clearing = False
        return

      vimsupport.PostVimMessage( text, warning = False, truncate = True )

    self._diag_message_needs_clearing = True


  def _DiagnosticsCount( self, predicate ):
    count = 0
    for diags in self._line_to_diags.values():
      count += sum( 1 for d in diags if predicate( d ) )
    return count


  def _UpdateLocationLists( self, open_on_edit = False ):
    vimsupport.SetLocationListsForBuffer(
      self._bufnr,
      vimsupport.ConvertDiagnosticsToQfList( self._diagnostics ),
      open_on_edit )


  def UpdateMatches( self ):
    if not self._user_options[ 'enable_diagnostic_highlighting' ]:
      return

    props_to_remove = vimsupport.GetTextProperties( self._bufnr )
    for diags in self._line_to_diags.values():
      # Insert squiggles in reverse order so that errors overlap warnings.
      for diag in reversed( diags ):
        for line, column, name, extras in _ConvertDiagnosticToTextProperties(
            self._bufnr,
            diag ):
          global YCM_VIM_PROPERTY_ID

          # Note the following .remove() works because the __eq__ on
          # DiagnosticProperty does not actually check the IDs match...
          diag_prop = vimsupport.DiagnosticProperty(
              YCM_VIM_PROPERTY_ID,
              name,
              line,
              column,
              extras[ 'end_col' ] - column if 'end_col' in extras else column )
          try:
            props_to_remove.remove( diag_prop )
          except ValueError:
            extras.update( {
              'id': YCM_VIM_PROPERTY_ID
            } )
            vimsupport.AddTextProperty( self._bufnr,
                                        line,
                                        column,
                                        name,
                                        extras )
          YCM_VIM_PROPERTY_ID += 1
    for prop in props_to_remove:
      vimsupport.RemoveDiagnosticProperty( self._bufnr, prop )


  def _UpdateSigns( self ):
    signs_to_unplace = vimsupport.GetSignsInBuffer( self._bufnr )
    signs_to_place = []
    for line, diags in self._line_to_diags.items():
      if not diags:
        continue

      # We always go for the first diagnostic on the line because diagnostics
      # are sorted by errors in priority and Vim can only display one sign by
      # line.
      name = 'YcmError' if _DiagnosticIsError( diags[ 0 ] ) else 'YcmWarning'
      sign = {
          'lnum': line,
          'name': name,
          'buffer': self._bufnr,
          'group': 'ycm_signs'
      }
      try:
        signs_to_unplace.remove( sign )
      except ValueError:
        signs_to_place.append( sign )
    vim.eval( f'sign_placelist( { signs_to_place } )' )
    vim.eval( f'sign_unplacelist( { signs_to_unplace } )' )


  def _ConvertDiagListToDict( self ):
    self._line_to_diags = defaultdict( list )
    for diag in self._diagnostics:
      location = diag[ 'location' ]
      bufnr = vimsupport.GetBufferNumberForFilename( location[ 'filepath' ] )
      if bufnr == self._bufnr:
        line_number = location[ 'line_num' ]
        self._line_to_diags[ line_number ].append( diag )

    for diags in self._line_to_diags.values():
      # We also want errors to be listed before warnings so that errors aren't
      # hidden by the warnings; Vim won't place a sign over an existing one.
      diags.sort( key = lambda diag: ( diag[ 'kind' ],
                                       diag[ 'location' ][ 'column_num' ] ) )


_DiagnosticIsError = CompileLevel( 'error' )
_DiagnosticIsWarning = CompileLevel( 'warning' )


def _NormalizeDiagnostic( diag ):
  def ClampToOne( value ):
    return value if value > 0 else 1

  location = diag[ 'location' ]
  location[ 'column_num' ] = ClampToOne( location[ 'column_num' ] )
  location[ 'line_num' ] = ClampToOne( location[ 'line_num' ] )
  return diag


def _ConvertDiagnosticToTextProperties( bufnr, diagnostic ):
  properties = []

  name = ( 'YcmErrorProperty' if _DiagnosticIsError( diagnostic ) else
            'YcmWarningProperty' )
  if vimsupport.VimIsNeovim():
    name = name.replace( 'Property', 'Section' )

  location_extent = diagnostic[ 'location_extent' ]
  if location_extent[ 'start' ][ 'line_num' ] <= 0:
    location = diagnostic[ 'location' ]
    line, column = vimsupport.LineAndColumnNumbersClamped(
      bufnr,
      location[ 'line_num' ],
      location[ 'column_num' ]
    )
    properties.append( ( line, column, name, {} ) )
  else:
    start_line, start_column = vimsupport.LineAndColumnNumbersClamped(
      bufnr,
      location_extent[ 'start' ][ 'line_num' ],
      location_extent[ 'start' ][ 'column_num' ]
    )
    end_line, end_column = vimsupport.LineAndColumnNumbersClamped(
      bufnr,
      location_extent[ 'end' ][ 'line_num' ],
      location_extent[ 'end' ][ 'column_num' ]
    )
    properties.append( (
      start_line,
      start_column,
      name,
      { 'end_lnum': end_line,
        'end_col': end_column } ) )

  for diagnostic_range in diagnostic[ 'ranges' ]:
    if ( diagnostic_range[ 'start' ][ 'line_num' ] == 0 or
         diagnostic_range[ 'end' ][ 'line_num' ] == 0 ):
      continue
    start_line, start_column = vimsupport.LineAndColumnNumbersClamped(
      bufnr,
      diagnostic_range[ 'start' ][ 'line_num' ],
      diagnostic_range[ 'start' ][ 'column_num' ]
    )
    end_line, end_column = vimsupport.LineAndColumnNumbersClamped(
      bufnr,
      diagnostic_range[ 'end' ][ 'line_num' ],
      diagnostic_range[ 'end' ][ 'column_num' ]
    )

    if not _IsValidRange( start_line, start_column, end_line, end_column ):
      continue

    properties.append( (
      start_line,
      start_column,
      name,
      { 'end_lnum': end_line,
        'end_col': end_column } ) )

  return properties


def _IsValidRange( start_line, start_column, end_line, end_column ):
  # End line before start line - invalid
  if start_line > end_line:
    return False

  # End line after start line - valid
  if start_line < end_line:
    return True

  # Same line, start colum after end column - invalid
  if start_column > end_column:
    return False

  # Same line, start column before or equal to end column - valid
  return True



# Copyright (C) 2015-2017 YouCompleteMe contributors.
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import vim
import re

# Can't import these from setup.py because it makes nosetests go crazy.
DIR_OF_CURRENT_SCRIPT = os.path.dirname( os.path.abspath( __file__ ) )
DIR_OF_YCMD = os.path.join( DIR_OF_CURRENT_SCRIPT, '..', '..', 'third_party',
                            'ycmd' )
WIN_PYTHON_PATH = os.path.join( sys.exec_prefix, 'python.exe' )
PYTHON_BINARY_REGEX = re.compile(
  r'python(3(\.[6-9])?)?(.exe)?$', re.IGNORECASE )


# Not caching the result of this function; users shouldn't have to restart Vim
# after running the install script or setting the
# `g:ycm_server_python_interpreter` option.
def PathToPythonInterpreter():
  # Not calling the Python interpreter to check its version as it significantly
  # impacts startup time.
  from ycmd import utils

  python_interpreter = vim.eval( 'g:ycm_server_python_interpreter' )
  if python_interpreter:
    python_interpreter = utils.FindExecutable( python_interpreter )
    if python_interpreter:
      return python_interpreter

    raise RuntimeError( "Path in 'g:ycm_server_python_interpreter' option "
                        "does not point to a valid Python 3.6+." )

  python_interpreter = _PathToPythonUsedDuringBuild()
  if python_interpreter and utils.GetExecutable( python_interpreter ):
    return python_interpreter

  # On UNIX platforms, we use sys.executable as the Python interpreter path.
  # We cannot use sys.executable on Windows because for unknown reasons, it
  # returns the Vim executable. Instead, we use sys.exec_prefix to deduce the
  # interpreter path.
  python_interpreter = ( WIN_PYTHON_PATH if utils.OnWindows() else
                         sys.executable )
  if _EndsWithPython( python_interpreter ):
    return python_interpreter

  python_interpreter = utils.PathToFirstExistingExecutable( [ 'python3',
                                                              'python' ] )
  if python_interpreter:
    return python_interpreter

  raise RuntimeError( "Cannot find Python 3.6+. "
                      "Set the 'g:ycm_server_python_interpreter' option "
                      "to a Python interpreter path." )


def _PathToPythonUsedDuringBuild():
  from ycmd import utils

  try:
    filepath = os.path.join( DIR_OF_YCMD, 'PYTHON_USED_DURING_BUILDING' )
    return utils.ReadFile( filepath ).strip()
  except OSError:
    return None


def _EndsWithPython( path ):
  """Check if given path ends with a python 3.6+ name."""
  return path and PYTHON_BINARY_REGEX.search( path ) is not None


def PathToServerScript():
  return os.path.join( DIR_OF_YCMD, 'ycmd' )



# Copyright (C) 2016 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.tests.test_utils import ( CurrentWorkingDirectory, ExtendedMock,
                                   MockVimModule, MockVimBuffers, VimBuffer )
MockVimModule()

import contextlib
from hamcrest import ( assert_that,
                       contains_exactly,
                       empty,
                       equal_to,
                       has_entries )
from unittest import TestCase
from unittest.mock import call, MagicMock, patch

from ycm.tests import PathToTestFile, YouCompleteMeInstance
from ycmd.responses import ServerError

import json


@contextlib.contextmanager
def MockCompletionRequest( response_method ):
  """Mock out the CompletionRequest, replacing the response handler
  JsonFromFuture with the |response_method| parameter."""

  # We don't want the requests to actually be sent to the server, just have it
  # return success.
  with patch( 'ycm.client.completer_available_request.'
              'CompleterAvailableRequest.PostDataToHandler',
              return_value = True ):
    with patch( 'ycm.client.completion_request.CompletionRequest.'
                'PostDataToHandlerAsync',
                return_value = MagicMock( return_value=True ) ):

      # We set up a fake response.
      with patch( 'ycm.client.base_request._JsonFromFuture',
                  side_effect = response_method ):
        yield


@contextlib.contextmanager
def MockResolveRequest( response_method ):
  """Mock out the CompletionRequest, replacing the response handler
  JsonFromFuture with the |response_method| parameter."""

  with patch( 'ycm.client.resolve_completion_request.ResolveCompletionRequest.'
              'PostDataToHandlerAsync',
              return_value = MagicMock( return_value=True ) ):

    # We set up a fake response.
    with patch( 'ycm.client.base_request._JsonFromFuture',
                side_effect = response_method ):
      yield


class CompletionTest( TestCase ):
  @YouCompleteMeInstance()
  def test_SendCompletionRequest_UnicodeWorkingDirectory( self, ycm ):
    unicode_dir = PathToTestFile( 'uni¢od€' )
    current_buffer = VimBuffer( PathToTestFile( 'uni¢𐍈d€', 'current_buffer' ) )

    def ServerResponse( *args ):
      return { 'completions': [], 'completion_start_column': 1 }

    with CurrentWorkingDirectory( unicode_dir ):
      with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
        with MockCompletionRequest( ServerResponse ):
          ycm.SendCompletionRequest()
          assert_that( ycm.CompletionRequestReady() )
          assert_that(
            ycm.GetCompletionResponse(),
            has_entries( {
              'completions': empty(),
              'completion_start_column': 1
            } )
          )


  @YouCompleteMeInstance()
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_SendCompletionRequest_ResponseContainingError(
      self, ycm, post_vim_message ):
    current_buffer = VimBuffer( 'buffer' )

    def ServerResponse( *args ):
      return {
        'completions': [ {
          'insertion_text': 'insertion_text',
          'menu_text': 'menu_text',
          'extra_menu_info': 'extra_menu_info',
          'detailed_info': 'detailed_info',
          'kind': 'kind',
          'extra_data': {
             'doc_string': 'doc_string'
          }
        } ],
        'completion_start_column': 3,
        'errors': [ {
          'exception': {
             'TYPE': 'Exception'
          },
          'message': 'message',
          'traceback': 'traceback'
        } ]
      }

    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with MockCompletionRequest( ServerResponse ):
        ycm.SendCompletionRequest()
        assert_that( ycm.CompletionRequestReady() )
        response = ycm.GetCompletionResponse()
        post_vim_message.assert_has_exact_calls( [
          call( 'Exception: message', truncate = True )
        ] )
        assert_that(
          response,
          has_entries( {
            'completions': contains_exactly( has_entries( {
              'word': 'insertion_text',
              'abbr': 'menu_text',
              'menu': 'extra_menu_info',
              'info': 'detailed_info\ndoc_string',
              'kind': 'k',
              'dup': 1,
              'empty': 1
            } ) ),
            'completion_start_column': 3
          } )
        )


  @YouCompleteMeInstance()
  @patch( 'ycm.client.base_request._logger', autospec = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_SendCompletionRequest_ErrorFromServer( self,
                                                  ycm,
                                                  post_vim_message,
                                                  logger ):
    current_buffer = VimBuffer( 'buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with MockCompletionRequest( ServerError( 'Server error' ) ):
        ycm.SendCompletionRequest()
        assert_that( ycm.CompletionRequestReady() )
        response = ycm.GetCompletionResponse()
        logger.exception.assert_called_with( 'Error while handling server '
                                             'response' )
        post_vim_message.assert_has_exact_calls( [
          call( 'Server error', truncate = True )
        ] )
        assert_that(
          response,
          has_entries( {
            'completions': empty(),
            'completion_start_column': -1
          } )
        )



  @YouCompleteMeInstance()
  @patch( 'ycm.client.base_request._logger', autospec = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_ResolveCompletionRequest_Resolves( self,
                                              ycm,
                                              post_vim_message,
                                              logger ):

    def CompletionResponse( *args ):
      return {
        'completions': [ {
          'insertion_text': 'insertion_text',
          'menu_text': 'menu_text',
          'extra_menu_info': 'extra_menu_info',
          'detailed_info': 'detailed_info',
          'kind': 'kind',
          'extra_data': {
            'doc_string': 'doc_string',
            'resolve': 10
          }
        } ],
        'completion_start_column': 3,
        'errors': []
      }

    def ResolveResponse( *args ):
      return {
        'completion': {
          'insertion_text': 'insertion_text',
          'menu_text': 'menu_text',
          'extra_menu_info': 'extra_menu_info',
          'detailed_info': 'detailed_info',
          'kind': 'kind',
          'extra_data': {
            'doc_string': 'doc_string with more info'
          }
        },
        'errors': []
      }

    current_buffer = VimBuffer( 'buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with MockCompletionRequest( CompletionResponse ):
        ycm.SendCompletionRequest()
        assert_that( ycm.CompletionRequestReady() )
        response = ycm.GetCompletionResponse()

        post_vim_message.assert_not_called()
        assert_that(
          response,
          has_entries( {
            'completions': contains_exactly( has_entries( {
              'word': 'insertion_text',
              'abbr': 'menu_text',
              'menu': 'extra_menu_info',
              'info': 'detailed_info\ndoc_string',
              'kind': 'k',
              'dup': 1,
              'empty': 1
            } ) ),
            'completion_start_column': 3
          } )
        )

        item = response[ 'completions' ][ 0 ]
        assert_that( json.loads( item[ 'user_data' ] ),
                     has_entries( { 'resolve': 10 } ) )

      with MockResolveRequest( ResolveResponse ):
        assert_that( ycm.ResolveCompletionItem( item ), equal_to( True ) )
        assert_that( ycm.CompletionRequestReady() )
        response = ycm.GetCompletionResponse()
        post_vim_message.assert_not_called()

        assert_that(
          response,
          has_entries( {
            'completion': has_entries( {
              'word': 'insertion_text',
              'abbr': 'menu_text',
              'menu': 'extra_menu_info',
              'info': 'detailed_info\ndoc_string with more info',
              'kind': 'k',
              'dup': 1,
              'empty': 1
            } )
          } )
        )

        item = response[ 'completion' ]

      with MockResolveRequest( ServerError( 'must not be called' ) ):
        assert_that( ycm.ResolveCompletionItem( item ), equal_to( False ) )
        post_vim_message.assert_not_called()


  @YouCompleteMeInstance()
  @patch( 'ycm.client.base_request._logger', autospec = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_ResolveCompletionRequest_ResponseContainsErrors( self,
                                                            ycm,
                                                            post_vim_message,
                                                            logger ):

    def CompletionResponse( *args ):
      return {
        'completions': [ {
          'insertion_text': 'insertion_text',
          'menu_text': 'menu_text',
          'extra_menu_info': 'extra_menu_info',
          'detailed_info': 'detailed_info',
          'kind': 'kind',
          'extra_data': {
            'doc_string': 'doc_string',
            'resolve': 10
          }
        } ],
        'completion_start_column': 3,
        'errors': []
      }

    def ResolveResponse( *args ):
      return {
        'completion': {
          'insertion_text': 'insertion_text',
          'menu_text': 'menu_text',
          'extra_menu_info': 'extra_menu_info',
          'detailed_info': 'detailed_info',
          'kind': 'kind',
          'extra_data': {
            'doc_string': 'doc_string with more info'
          }
        },
        'errors': [ {
          'exception': {
             'TYPE': 'Exception'
          },
          'message': 'message',
          'traceback': 'traceback'
        } ]
      }

    current_buffer = VimBuffer( 'buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with MockCompletionRequest( CompletionResponse ):
        ycm.SendCompletionRequest()
        assert_that( ycm.CompletionRequestReady() )
        response = ycm.GetCompletionResponse()

        post_vim_message.assert_not_called()
        assert_that(
          response,
          has_entries( {
            'completions': contains_exactly( has_entries( {
              'word': 'insertion_text',
              'abbr': 'menu_text',
              'menu': 'extra_menu_info',
              'info': 'detailed_info\ndoc_string',
              'kind': 'k',
              'dup': 1,
              'empty': 1
            } ) ),
            'completion_start_column': 3
          } )
        )

        item = response[ 'completions' ][ 0 ]
        assert_that( json.loads( item[ 'user_data' ] ),
                     has_entries( { 'resolve': 10 } ) )

      with MockResolveRequest( ResolveResponse ):
        assert_that( ycm.ResolveCompletionItem( item ), equal_to( True ) )
        assert_that( ycm.CompletionRequestReady() )
        response = ycm.GetCompletionResponse()

        post_vim_message.assert_has_exact_calls( [
          call( 'Exception: message', truncate = True )
        ] )
        assert_that(
          response,
          has_entries( {
            'completion': has_entries( {
              'word': 'insertion_text',
              'abbr': 'menu_text',
              'menu': 'extra_menu_info',
              'info': 'detailed_info\ndoc_string with more info',
              'kind': 'k',
              'dup': 1,
              'empty': 1
            } )
          } )
        )


  @YouCompleteMeInstance()
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_ResolveCompletionItem_NoUserData( self, ycm, post_vim_message ):
    def CompletionResponse( *args ):
      return {
        'completions': [ {
          'insertion_text': 'insertion_text',
          'menu_text': 'menu_text',
          'extra_menu_info': 'extra_menu_info',
          'detailed_info': 'detailed_info',
          'kind': 'kind'
        } ],
        'completion_start_column': 3,
        'errors': []
      }

    current_buffer = VimBuffer( 'buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with MockCompletionRequest( CompletionResponse ):
        ycm.SendCompletionRequest()
        assert_that( ycm.CompletionRequestReady() )
        response = ycm.GetCompletionResponse()

        post_vim_message.assert_not_called()
        assert_that(
          response,
          has_entries( {
            'completions': contains_exactly( has_entries( {
              'word': 'insertion_text',
              'abbr': 'menu_text',
              'menu': 'extra_menu_info',
              'info': 'detailed_info',
              'kind': 'k',
              'dup': 1,
              'empty': 1
            } ) ),
            'completion_start_column': 3
          } )
        )

        item = response[ 'completions' ][ 0 ]
        item.pop( 'user_data' )

      with MockResolveRequest( ServerError( 'must not be called' ) ):
        assert_that( ycm.ResolveCompletionItem( item ), equal_to( False ) )
        post_vim_message.assert_not_called()


  @YouCompleteMeInstance()
  def test_ResolveCompletionItem_NoRequest( self, ycm ):
    assert_that( ycm.GetCurrentCompletionRequest(), equal_to( None ) )
    assert_that( ycm.ResolveCompletionItem( {} ), equal_to( False ) )


  @YouCompleteMeInstance()
  @patch( 'ycm.client.base_request._logger', autospec = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_ResolveCompletionRequest_ServerError(
      self, ycm, post_vim_message, logger ):

    def ServerResponse( *args ):
      return {
        'completions': [ {
          'insertion_text': 'insertion_text',
          'menu_text': 'menu_text',
          'extra_menu_info': 'extra_menu_info',
          'detailed_info': 'detailed_info',
          'kind': 'kind',
          'extra_data': {
            'doc_string': 'doc_string',
            'resolve': 10
          }
        } ],
        'completion_start_column': 3,
        'errors': []
      }

    current_buffer = VimBuffer( 'buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with MockCompletionRequest( ServerResponse ):
        ycm.SendCompletionRequest()
        assert_that( ycm.CompletionRequestReady() )
        response = ycm.GetCompletionResponse()

        post_vim_message.assert_not_called()
        assert_that(
          response,
          has_entries( {
            'completions': contains_exactly( has_entries( {
              'word': 'insertion_text',
              'abbr': 'menu_text',
              'menu': 'extra_menu_info',
              'info': 'detailed_info\ndoc_string',
              'kind': 'k',
              'dup': 1,
              'empty': 1
            } ) ),
            'completion_start_column': 3
          } )
        )

        item = response[ 'completions' ][ 0 ]
        assert_that( json.loads( item[ 'user_data' ] ),
                     has_entries( { 'resolve': 10 } ) )

      with MockResolveRequest( ServerError( 'Server error' ) ):
        ycm.ResolveCompletionItem( item )
        assert_that( ycm.CompletionRequestReady() )
        response = ycm.GetCompletionResponse()

        logger.exception.assert_called_with( 'Error while handling server '
                                             'response' )
        post_vim_message.assert_has_exact_calls( [
          call( 'Server error', truncate = True )
        ] )



# Copyright (C) 2013-2018 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.client.base_request import BaseRequest, BuildRequestData


class EventNotification( BaseRequest ):
  def __init__( self, event_name, buffer_number = None, extra_data = None ):
    super( EventNotification, self ).__init__()
    self._event_name = event_name
    self._buffer_number = buffer_number
    self._extra_data = extra_data
    self._response_future = None
    self._cached_response = None


  def Start( self ):
    request_data = BuildRequestData( self._buffer_number )
    if self._extra_data:
      request_data.update( self._extra_data )
    request_data[ 'event_name' ] = self._event_name

    self._response_future = self.PostDataToHandlerAsync( request_data,
                                                         'event_notification' )


  def Done( self ):
    return bool( self._response_future ) and self._response_future.done()


  def Response( self ):
    if self._cached_response:
      return self._cached_response

    if not self._response_future or self._event_name != 'FileReadyToParse':
      return []

    self._cached_response = self.HandleFuture( self._response_future,
                                               truncate_message = True )

    return self._cached_response if self._cached_response else []


def SendEventNotificationAsync( event_name,
                                buffer_number = None,
                                extra_data = None ):
  event = EventNotification( event_name, buffer_number, extra_data )
  event.Start()



# Copyright (C) 2016-2018 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.tests.test_utils import MockVimModule, MockVimBuffers, VimBuffer
MockVimModule()

from hamcrest import assert_that, contains_exactly, has_entries
from unittest.mock import patch
from unittest import TestCase

from ycm.tests import YouCompleteMeInstance


class CommandTest( TestCase ):
  @YouCompleteMeInstance( { 'g:ycm_extra_conf_vim_data': [ 'tempname()' ] } )
  def test_SendCommandRequest_ExtraConfVimData_Works( self, ycm ):
    current_buffer = VimBuffer( 'buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with patch( 'ycm.youcompleteme.SendCommandRequest' ) as send_request:
        ycm.SendCommandRequest( [ 'GoTo' ], 'aboveleft', False, 1, 1 )
        assert_that(
          # Positional arguments passed to SendCommandRequest.
          send_request.call_args[ 0 ],
          contains_exactly(
            contains_exactly( 'GoTo' ),
            'aboveleft',
            'same-buffer',
            has_entries( {
              'options': has_entries( {
                'tab_size': 2,
                'insert_spaces': True,
              } ),
              'extra_conf_data': has_entries( {
                'tempname()': '_TEMP_FILE_'
              } ),
            } )
          )
        )


  @YouCompleteMeInstance( {
    'g:ycm_extra_conf_vim_data': [ 'undefined_value' ] } )
  def test_SendCommandRequest_ExtraConfData_UndefinedValue( self, ycm ):
    current_buffer = VimBuffer( 'buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with patch( 'ycm.youcompleteme.SendCommandRequest' ) as send_request:
        ycm.SendCommandRequest( [ 'GoTo' ], 'belowright', False, 1, 1 )
        assert_that(
          # Positional arguments passed to SendCommandRequest.
          send_request.call_args[ 0 ],
          contains_exactly(
            contains_exactly( 'GoTo' ),
            'belowright',
            'same-buffer',
            has_entries( {
              'options': has_entries( {
                'tab_size': 2,
                'insert_spaces': True,
              } )
            } )
          )
        )


  @YouCompleteMeInstance()
  def test_SendCommandRequest_BuildRange_NoVisualMarks( self, ycm, *args ):
    current_buffer = VimBuffer( 'buffer', contents = [ 'first line',
                                                       'second line' ] )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with patch( 'ycm.youcompleteme.SendCommandRequest' ) as send_request:
        ycm.SendCommandRequest( [ 'GoTo' ], '', True, 1, 2 )
        send_request.assert_called_once_with(
          [ 'GoTo' ],
          '',
          'same-buffer',
          {
            'options': {
              'tab_size': 2,
              'insert_spaces': True
            },
            'range': {
              'start': {
                'line_num': 1,
                'column_num': 1
              },
              'end': {
                'line_num': 2,
                'column_num': 12
              }
            }
          }
        )


  @YouCompleteMeInstance()
  def test_SendCommandRequest_BuildRange_VisualMarks( self, ycm, *args ):
    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'first line',
                                             'second line' ],
                                visual_start = [ 1, 4 ],
                                visual_end = [ 2, 8 ] )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with patch( 'ycm.youcompleteme.SendCommandRequest' ) as send_request:
        ycm.SendCommandRequest( [ 'GoTo' ], 'tab', True, 1, 2 )
        send_request.assert_called_once_with(
          [ 'GoTo' ],
          'tab',
          'same-buffer',
          {
            'options': {
              'tab_size': 2,
              'insert_spaces': True
            },
            'range': {
              'start': {
                'line_num': 1,
                'column_num': 5
              },
              'end': {
                'line_num': 2,
                'column_num': 9
              }
            }
          }
        )


  @YouCompleteMeInstance()
  def test_SendCommandRequest_IgnoreFileTypeOption( self, ycm, *args ):
    current_buffer = VimBuffer( 'buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      expected_args = (
        [ 'GoTo' ],
        '',
        'same-buffer',
        {
          'completer_target': 'python',
          'options': {
            'tab_size': 2,
            'insert_spaces': True
          },
        }
      )

      with patch( 'ycm.youcompleteme.SendCommandRequest' ) as send_request:
        ycm.SendCommandRequest( [ 'ft=python', 'GoTo' ], '', False, 1, 1 )
        send_request.assert_called_once_with( *expected_args )

      with patch( 'ycm.youcompleteme.SendCommandRequest' ) as send_request:
        ycm.SendCommandRequest( [ 'GoTo', 'ft=python' ], '', False, 1, 1 )
        send_request.assert_called_once_with( *expected_args )



# Copyright (C) 2022, YouCompleteMe Contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.


from ycm.client.inlay_hints_request import InlayHintsRequest
from ycm.client.base_request import BuildRequestData
from ycm import vimsupport
from ycm import text_properties as tp


HIGHLIGHT_GROUP = {
  'Type':      'YcmInlayHint',
  'Parameter': 'YcmInlayHint',
  'Enum':      'YcmInlayHint',
}
REPORTED_MISSING_TYPES = set()


def Initialise():
  if not vimsupport.VimSupportsVirtualText():
    return False

  props = tp.GetTextPropertyTypes()
  if 'YCM_INLAY_UNKNOWN' not in props:
    tp.AddTextPropertyType( 'YCM_INLAY_UNKNOWN',
                            highlight = 'YcmInlayHint',
                            start_incl = 1 )
  if 'YCM_INLAY_PADDING' not in props:
    tp.AddTextPropertyType( 'YCM_INLAY_PADDING',
                            highlight = 'YcmInvisible',
                            start_incl = 1 )

  for token_type, group in HIGHLIGHT_GROUP.items():
    prop = f'YCM_INLAY_{ token_type }'
    if prop not in props and vimsupport.GetIntValue(
        f"hlexists( '{ vimsupport.EscapeForVim( group ) }' )" ):
      tp.AddTextPropertyType( prop,
                              highlight = group,
                              start_incl = 1 )

  return True


class InlayHints:
  """Stores the inlay hints state for a Vim buffer"""

  # FIXME: Send a request per-disjoint range for this buffer rather than the
  # maximal range. then collaate the results when all responses are returned
  def __init__( self, bufnr, user_options ):
    self._request = None
    self._bufnr = bufnr
    self.tick = -1
    self._latest_inlay_hints = []
    self._last_requested_range = None


  def Request( self, force=False ):
    if self._request and not self.Ready():
      return True

    # Check to see if the buffer ranges would actually change anything visible.
    # This avoids a round-trip for every single line scroll event
    if ( not force and
         self.tick == vimsupport.GetBufferChangedTick( self._bufnr ) and
         vimsupport.VisibleRangeOfBufferOverlaps(
           self._bufnr,
           self._last_requested_range ) ):
      return False # don't poll

    # We're requesting changes, so the existing results are now invalid
    self._latest_inlay_hints = []
    # FIXME: This call is duplicated in the call to VisibleRangeOfBufferOverlaps
    #  - remove the expansion param
    #  - look up the actual visible range, then call this function
    #  - if not overlapping, do the factor expansion and request
    self._last_requested_range = vimsupport.RangeVisibleInBuffer( self._bufnr )
    self.tick = vimsupport.GetBufferChangedTick( self._bufnr )

    request_data = BuildRequestData( self._bufnr )
    request_data.update( {
      'range': self._last_requested_range
    } )
    self._request = InlayHintsRequest( request_data )
    self._request.Start()
    return True


  def Ready( self ):
    return self._request is not None and self._request.Done()


  def Clear( self ):
    # ClearTextProperties is slow as it must scan the whole buffer
    # we shouldn't use _last_requested_range, because the server is free to
    # return a larger range, so we pick the first/last from the latest results
    types = [ 'YCM_INLAY_UNKNOWN', 'YCM_INLAY_PADDING' ] + [
      f'YCM_INLAY_{ prop_type }' for prop_type in HIGHLIGHT_GROUP.keys()
    ]

    tp.ClearTextProperties( self._bufnr, prop_types = types )

  def Update( self ):
    if not self._request:
      # Nothing to update
      return True

    assert self.Ready()

    # We're ready to use this response. Clear it (to avoid repeatedly
    # re-polling).
    self._latest_inlay_hints = self._request.Response()
    self._request = None

    if self.tick != vimsupport.GetBufferChangedTick( self._bufnr ):
      # Buffer has changed, we should ignore the data and retry
      self.Request( force=True )
      return False # poll again

    self._Draw()

    # No need to re-poll
    return True


  def Refresh( self ):
    if self.tick != vimsupport.GetBufferChangedTick( self._bufnr ):
      # stale data
      return

    if self._request is not None:
      # request in progress; we''l handle refreshing when it's done.
      return

    self._Draw()


  def _Draw( self ):
    self.Clear()

    for inlay_hint in self._latest_inlay_hints:
      if 'kind' not in inlay_hint:
        prop_type = 'YCM_INLAY_UNKNOWN'
      elif inlay_hint[ 'kind' ] not in HIGHLIGHT_GROUP:
        prop_type = 'YCM_INLAY_UNKNOWN'
      else:
        prop_type = 'YCM_INLAY_' + inlay_hint[ 'kind' ]

      if inlay_hint.get( 'paddingLeft', False ):
        tp.AddTextProperty( self._bufnr,
                            None,
                            'YCM_INLAY_PADDING',
                            {
                              'start': inlay_hint[ 'position' ],
                            },
                            {
                              'text': ' '
                            } )

      tp.AddTextProperty( self._bufnr,
                          None,
                          prop_type,
                          {
                            'start': inlay_hint[ 'position' ],
                          },
                          {
                            'text': inlay_hint[ 'label' ]
                          } )

      if inlay_hint.get( 'paddingRight', False ):
        tp.AddTextProperty( self._bufnr,
                            None,
                            'YCM_INLAY_PADDING',
                            {
                              'start': inlay_hint[ 'position' ],
                            },
                            {
                              'text': ' '
                            } )



# encoding: utf-8
#
# Copyright (C) 2015-2016 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.tests.test_utils import MockVimModule
MockVimModule()

import contextlib
import json
from hamcrest import assert_that, contains_exactly, empty, equal_to, none
from unittest import TestCase
from unittest.mock import MagicMock, DEFAULT, patch

from ycm import vimsupport
from ycmd.utils import ToBytes
from ycm.client.completion_request import ( CompletionRequest,
                                            _FilterToMatchingCompletions,
                                            _GetRequiredNamespaceImport )
from ycm.client.omni_completion_request import OmniCompletionRequest


def CompleteItemIs( word, abbr = None, menu = None,
                    info = None, kind = None, **kwargs ):
  item = {
    'word': ToBytes( word ),
    'abbr': ToBytes( abbr ),
    'menu': ToBytes( menu ),
    'info': ToBytes( info ),
    'kind': ToBytes( kind ),
  }
  item.update( **kwargs )
  return item


def GetVariableValue_CompleteItemIs( word, abbr = None, menu = None,
                                     info = None, kind = None, **kwargs ):
  def Result( variable ):
    if variable == 'v:completed_item':
      return CompleteItemIs( word, abbr, menu, info, kind, **kwargs )
    return DEFAULT
  return MagicMock( side_effect = Result )


def BuildCompletion( insertion_text = 'Test',
                     menu_text = None,
                     extra_menu_info = None,
                     detailed_info = None,
                     kind = None,
                     extra_data = None ):
  completion = {
    'insertion_text': insertion_text
  }

  if extra_menu_info:
    completion[ 'extra_menu_info' ] = extra_menu_info
  if menu_text:
    completion[ 'menu_text' ] = menu_text
  if detailed_info:
    completion[ 'detailed_info' ] = detailed_info
  if kind:
    completion[ 'kind' ] = kind
  if extra_data:
    completion[ 'extra_data' ] = extra_data
  return completion


def BuildCompletionNamespace( namespace = None,
                              insertion_text = 'Test',
                              menu_text = None,
                              extra_menu_info = None,
                              detailed_info = None,
                              kind = None ):
  return BuildCompletion( insertion_text = insertion_text,
                          menu_text = menu_text,
                          extra_menu_info = extra_menu_info,
                          detailed_info = detailed_info,
                          kind = kind,
                          extra_data = {
                            'required_namespace_import': namespace
                          } )


def BuildCompletionFixIt( fixits,
                          insertion_text = 'Test',
                          menu_text = None,
                          extra_menu_info = None,
                          detailed_info = None,
                          kind = None ):
  return BuildCompletion( insertion_text = insertion_text,
                          menu_text = menu_text,
                          extra_menu_info = extra_menu_info,
                          detailed_info = detailed_info,
                          kind = kind,
                          extra_data = {
                            'fixits': fixits,
                          } )


@contextlib.contextmanager
def _SetupForCsharpCompletionDone( completions ):
  with patch( 'ycm.vimsupport.InsertNamespace' ):
    with _SetUpCompleteDone( completions ) as request:
      yield request


@contextlib.contextmanager
def _SetUpCompleteDone( completions ):
  with patch( 'ycm.vimsupport.TextBeforeCursor', return_value = '   Test' ):
    request = CompletionRequest( None )
    request.Done = MagicMock( return_value = True )
    request._RawResponse = MagicMock( return_value = {
      'completions': completions
    } )
    yield request


class PostcompleteTest( TestCase ):
  @patch( 'ycm.vimsupport.CurrentFiletypes', return_value = [ 'ycmtest' ] )
  def test_OnCompleteDone_DefaultFixIt( self, *args ):
    request = CompletionRequest( None )
    request.Done = MagicMock( return_value = True )
    request._OnCompleteDone_Csharp = MagicMock()
    request._OnCompleteDone_FixIt = MagicMock()
    request.OnCompleteDone()
    request._OnCompleteDone_Csharp.assert_not_called()
    request._OnCompleteDone_FixIt.assert_called_once_with()


  @patch( 'ycm.vimsupport.CurrentFiletypes', return_value = [ 'cs' ] )
  def test_OnCompleteDone_CsharpFixIt( self, *args ):
    request = CompletionRequest( None )
    request.Done = MagicMock( return_value = True )
    request._OnCompleteDone_Csharp = MagicMock()
    request._OnCompleteDone_FixIt = MagicMock()
    request.OnCompleteDone()
    request._OnCompleteDone_Csharp.assert_called_once_with()
    request._OnCompleteDone_FixIt.assert_not_called()


  @patch( 'ycm.vimsupport.CurrentFiletypes', return_value = [ 'ycmtest' ] )
  def test_OnCompleteDone_NoFixItIfNotDone( self, *args ):
    request = CompletionRequest( None )
    request.Done = MagicMock( return_value = False )
    request._OnCompleteDone_Csharp = MagicMock()
    request._OnCompleteDone_FixIt = MagicMock()
    request.OnCompleteDone()
    request._OnCompleteDone_Csharp.assert_not_called()
    request._OnCompleteDone_FixIt.assert_not_called()


  @patch( 'ycm.vimsupport.CurrentFiletypes', return_value = [ 'ycmtest' ] )
  def test_OnCompleteDone_NoFixItForOmnifunc( self, *args ):
    request = OmniCompletionRequest( 'omnifunc', None )
    request.Done = MagicMock( return_value = True )
    request._OnCompleteDone_Csharp = MagicMock()
    request._OnCompleteDone_FixIt = MagicMock()
    request.OnCompleteDone()
    request._OnCompleteDone_Csharp.assert_not_called()
    request._OnCompleteDone_FixIt.assert_not_called()


  def test_FilterToCompletedCompletions_MatchIsReturned( self ):
    completions = [ BuildCompletion( insertion_text = 'Test' ) ]
    result = _FilterToMatchingCompletions( CompleteItemIs( 'Test' ),
                                           completions )
    assert_that( list( result ), contains_exactly( {} ) )


  def test_FilterToCompletedCompletions_ShortTextDoesntRaise( self ):
    completions = [ BuildCompletion( insertion_text = 'AAA' ) ]
    result = _FilterToMatchingCompletions( CompleteItemIs( 'A' ), completions )
    assert_that( list( result ), empty() )


  def test_FilterToCompletedCompletions_ExactMatchIsReturned( self ):
    completions = [ BuildCompletion( insertion_text = 'Test' ) ]
    result = _FilterToMatchingCompletions( CompleteItemIs( 'Test' ),
                                           completions )
    assert_that( list( result ), contains_exactly( {} ) )


  def test_FilterToCompletedCompletions_NonMatchIsntReturned( self ):
    completions = [ BuildCompletion( insertion_text = 'A' ) ]
    result = _FilterToMatchingCompletions( CompleteItemIs( '   Quote' ),
                                           completions )
    assert_that( list( result ), empty() )


  def test_FilterToCompletedCompletions_Unicode( self ):
    completions = [ BuildCompletion( insertion_text = '†es†' ) ]
    result = _FilterToMatchingCompletions( CompleteItemIs( '†es†' ),
                                           completions )
    assert_that( list( result ), contains_exactly( {} ) )


  def test_GetRequiredNamespaceImport_ReturnNoneForNoExtraData( self ):
    assert_that( _GetRequiredNamespaceImport( {} ), none() )


  def test_GetRequiredNamespaceImport_ReturnNamespaceFromExtraData( self ):
    namespace = 'A_NAMESPACE'
    assert_that( _GetRequiredNamespaceImport(
                     BuildCompletionNamespace( namespace )[ 'extra_data' ] ),
                 equal_to( namespace ) )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs( 'Te' ) )
  def test_GetExtraDataUserMayHaveCompleted_ReturnEmptyIfPendingMatches(
      *args ):
    completions = [ BuildCompletionNamespace( None ) ]
    with _SetupForCsharpCompletionDone( completions ) as request:
      assert_that( request._GetExtraDataUserMayHaveCompleted(), empty() )


  def test_GetExtraDataUserMayHaveCompleted_ReturnMatchIfExactMatches(
      self, *args ):
    info = [ 'NS', 'Test', 'Abbr', 'Menu', 'Info', 'Kind' ]
    completions = [ BuildCompletionNamespace( *info ) ]
    with _SetupForCsharpCompletionDone( completions ) as request:
      with patch( 'ycm.vimsupport.GetVariableValue',
                  GetVariableValue_CompleteItemIs( *info[ 1: ] ) ):
        assert_that( request._GetExtraDataUserMayHaveCompleted(),
                     contains_exactly( completions[ 0 ][ 'extra_data' ] ) )


  def test_GetExtraDataUserMayHaveCompleted_ReturnMatchIfExactMatchesEvenIfPartial( self ): # noqa
    info = [ 'NS', 'Test', 'Abbr', 'Menu', 'Info', 'Kind' ]
    completions = [ BuildCompletionNamespace( *info ),
                    BuildCompletion( insertion_text = 'TestTest' ) ]
    with _SetupForCsharpCompletionDone( completions ) as request:
      with patch( 'ycm.vimsupport.GetVariableValue',
                  GetVariableValue_CompleteItemIs( *info[ 1: ] ) ):
        assert_that( request._GetExtraDataUserMayHaveCompleted(),
                     contains_exactly( completions[ 0 ][ 'extra_data' ] ) )


  def test_GetExtraDataUserMayHaveCompleted_DontReturnMatchIfNoExactMatchesAndPartial( self ): # noqa
    info = [ 'NS', 'Test', 'Abbr', 'Menu', 'Info', 'Kind' ]
    completions = [ BuildCompletion( insertion_text = info[ 0 ] ),
                    BuildCompletion( insertion_text = 'TestTest' ) ]
    with _SetupForCsharpCompletionDone( completions ) as request:
      with patch( 'ycm.vimsupport.GetVariableValue',
                  GetVariableValue_CompleteItemIs( *info[ 1: ] ) ):
        assert_that( request._GetExtraDataUserMayHaveCompleted(), empty() )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs( 'Test' ) )
  def test_GetExtraDataUserMayHaveCompleted_ReturnMatchIfMatches( self, *args ):
    completions = [ BuildCompletionNamespace( None ) ]
    with _SetupForCsharpCompletionDone( completions ) as request:
      assert_that( request._GetExtraDataUserMayHaveCompleted(),
                   contains_exactly( completions[ 0 ][ 'extra_data' ] ) )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs(
            'Test',
            user_data=json.dumps( {
              'required_namespace_import': 'namespace1' } ) ) )
  def test_GetExtraDataUserMayHaveCompleted_UseUserData0( self, *args ):
    # Identical completions but we specify the first one via user_data.
    completions = [
      BuildCompletionNamespace( 'namespace1' ),
      BuildCompletionNamespace( 'namespace2' )
    ]

    with _SetupForCsharpCompletionDone( completions ) as request:
      assert_that(
          request._GetExtraDataUserMayHaveCompleted(),
          contains_exactly(
            BuildCompletionNamespace( 'namespace1' )[ 'extra_data' ] ) )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs(
            'Test',
            user_data=json.dumps( {
              'required_namespace_import': 'namespace2' } ) ) )
  def test_GetExtraDataUserMayHaveCompleted_UseUserData1( self, *args ):
    # Identical completions but we specify the second one via user_data.
    completions = [
      BuildCompletionNamespace( 'namespace1' ),
      BuildCompletionNamespace( 'namespace2' )
    ]

    with _SetupForCsharpCompletionDone( completions ) as request:
      assert_that(
          request._GetExtraDataUserMayHaveCompleted(),
          contains_exactly(
            BuildCompletionNamespace( 'namespace2' )[ 'extra_data' ] ) )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs( 'Test', user_data='' ) )
  def test_GetExtraDataUserMayHaveCompleted_EmptyUserData( self, *args ):
    # Identical completions but none is selected.
    completions = [
      BuildCompletionNamespace( 'namespace1' ),
      BuildCompletionNamespace( 'namespace2' )
    ]

    with _SetupForCsharpCompletionDone( completions ) as request:
      assert_that( request._GetExtraDataUserMayHaveCompleted(), empty() )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs( 'Test' ) )
  def test_PostCompleteCsharp_EmptyDoesntInsertNamespace( self, *args ):
    with _SetupForCsharpCompletionDone( [] ) as request:
      request._OnCompleteDone_Csharp()
      assert_that( not vimsupport.InsertNamespace.called )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs( 'Test' ) )
  def test_PostCompleteCsharp_ExistingWithoutNamespaceDoesntInsertNamespace(
      self, *args ):
    completions = [ BuildCompletionNamespace( None ) ]
    with _SetupForCsharpCompletionDone( completions ) as request:
      request._OnCompleteDone_Csharp()
      assert_that( not vimsupport.InsertNamespace.called )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs( 'Test' ) )
  def test_PostCompleteCsharp_ValueDoesInsertNamespace( self, *args ):
    namespace = 'A_NAMESPACE'
    completions = [ BuildCompletionNamespace( namespace ) ]
    with _SetupForCsharpCompletionDone( completions ) as request:
      request._OnCompleteDone_Csharp()
      vimsupport.InsertNamespace.assert_called_once_with( namespace )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs( 'Test' ) )
  @patch( 'ycm.vimsupport.PresentDialog', return_value = 1 )
  def test_PostCompleteCsharp_InsertSecondNamespaceIfSelected( self, *args ):
    namespace = 'A_NAMESPACE'
    namespace2 = 'ANOTHER_NAMESPACE'
    completions = [
      BuildCompletionNamespace( namespace ),
      BuildCompletionNamespace( namespace2 ),
    ]
    with _SetupForCsharpCompletionDone( completions ) as request:
      request._OnCompleteDone_Csharp()
      vimsupport.InsertNamespace.assert_called_once_with( namespace2 )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs( 'Test' ) )
  @patch( 'ycm.vimsupport.ReplaceChunks' )
  def test_PostCompleteFixIt_ApplyFixIt_NoFixIts( self, replace_chunks, *args ):
    completions = [
      BuildCompletionFixIt( [] )
    ]
    with _SetUpCompleteDone( completions ) as request:
      request._OnCompleteDone_FixIt()
      replace_chunks.assert_not_called()


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs( 'Test' ) )
  @patch( 'ycm.vimsupport.ReplaceChunks' )
  def test_PostCompleteFixIt_ApplyFixIt_EmptyFixIt(
      self, replace_chunks, *args ):
    completions = [
      BuildCompletionFixIt( [ { 'chunks': [] } ] )
    ]
    with _SetUpCompleteDone( completions ) as request:
      request._OnCompleteDone_FixIt()
      replace_chunks.assert_called_once_with( [], silent = True )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs( 'Test' ) )
  @patch( 'ycm.vimsupport.ReplaceChunks' )
  def test_PostCompleteFixIt_ApplyFixIt_NoFixIt( self, replace_chunks, *args ):
    completions = [
      BuildCompletion()
    ]
    with _SetUpCompleteDone( completions ) as request:
      request._OnCompleteDone_FixIt()
      replace_chunks.assert_not_called()


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs( 'Test' ) )
  @patch( 'ycm.vimsupport.ReplaceChunks' )
  def test_PostCompleteFixIt_ApplyFixIt_PickFirst(
      self, replace_chunks, *args ):
    completions = [
      BuildCompletionFixIt( [ { 'chunks': 'one' } ] ),
      BuildCompletionFixIt( [ { 'chunks': 'two' } ] ),
    ]
    with _SetUpCompleteDone( completions ) as request:
      request._OnCompleteDone_FixIt()
      replace_chunks.assert_called_once_with( 'one', silent = True )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs(
            'Test',
            user_data=json.dumps( { 'fixits': [ { 'chunks': 'one' } ] } ) ) )
  @patch( 'ycm.vimsupport.ReplaceChunks' )
  def test_PostCompleteFixIt_ApplyFixIt_PickFirstUserData( self,
                                                           replace_chunks,
                                                           *args ):
    completions = [
      BuildCompletionFixIt( [ { 'chunks': 'one' } ] ),
      BuildCompletionFixIt( [ { 'chunks': 'two' } ] ),
    ]
    with _SetUpCompleteDone( completions ) as request:
      request._OnCompleteDone_FixIt()
      replace_chunks.assert_called_once_with( 'one', silent = True )


  @patch( 'ycm.vimsupport.GetVariableValue',
          GetVariableValue_CompleteItemIs(
            'Test',
            user_data=json.dumps( { 'fixits': [ { 'chunks': 'two' } ] } ) ) )
  @patch( 'ycm.vimsupport.ReplaceChunks' )
  def test_PostCompleteFixIt_ApplyFixIt_PickSecond(
      self, replace_chunks, *args ):
    completions = [
      BuildCompletionFixIt( [ { 'chunks': 'one' } ] ),
      BuildCompletionFixIt( [ { 'chunks': 'two' } ] ),
    ]
    with _SetUpCompleteDone( completions ) as request:
      request._OnCompleteDone_FixIt()
      replace_chunks.assert_called_once_with( 'two', silent = True )
